@@ -1,8 +1,8 @@
 package Unicode::Collate;
 
 BEGIN {
-    if (ord("A") == 193) {
-	die "Unicode::Collate not ported to EBCDIC\n";
+    unless ("A" eq pack('U', 0x41) || "A" eq pack('U', ord("A"))) {
+	die "Unicode::Collate cannot stringify a Unicode code point\n";
     }
 }
 
@@ -14,11 +14,7 @@ use File::Spec;
 
 require Exporter;
 
-# Supporting on EBCDIC platform is not tested.
-# Tester(s) welcome!
-our $IsEBCDIC = ord("A") != 0x41;
-
-our $VERSION = '0.22';
+our $VERSION = '0.24';
 our $PACKAGE = __PACKAGE__;
 
 our @ISA = qw(Exporter);
@@ -37,7 +33,7 @@ eval { require Unicode::UCD };
 unless ($@) {
     $UNICODE_VERSION = Unicode::UCD::UnicodeVersion();
 }
-else { # XXX, Perl 5.6.1
+else { # Perl 5.6.1
     my($f, $fh);
     foreach my $d (@INC) {
 	$f = File::Spec->catfile($d, "unicode", "Unicode.301");
@@ -59,24 +55,30 @@ use constant NOMATCHPOS => -1;
 # This is also used as a HAS_UNICODE_NORMALIZE flag.
 our $getCombinClass;
 
-# minimum weight at level 2 and 3, respectively
-use constant Min2   => 0x20;
-use constant Min3   => 0x02;
+# Supported Levels
+use constant MinLevel => 1;
+use constant MaxLevel => 4;
 
-# shifted weight at 4th level
-use constant Shift4 => 0xFFFF;
+# Minimum weights at level 2 and 3, respectively
+use constant Min2Wt => 0x20;
+use constant Min3Wt => 0x02;
 
-# a boolean for Variable and 16-bit weights at 4 levels of Collation Element
+# Shifted weight at 4th level
+use constant Shift4Wt => 0xFFFF;
+
+# Variable weight at 1st level.
+# This is a negative value but should be regarded as zero on collation.
+# This is for distinction of variable chars from level 3 ignorable chars.
+use constant Var1Wt => -1;
+
+
+# A boolean for Variable and 16-bit weights at 4 levels of Collation Element
 # PROBLEM: The Default Unicode Collation Element Table
 # has weights over 0xFFFF at the 4th level.
 # The tie-breaking in the variable weights
 # other than "shift" (as well as "shift-trimmed") is unreliable.
 use constant VCE_TEMPLATE => 'Cn4';
 
-# Unicode encoding of strings to be collated
-# TODO: 'N*' for UTF-32BE, 'V*' for UTF-32LE.
-use constant UTF_TEMPLATE => 'U*';
-
 # A sort key: 16-bit weights
 # See also the PROBLEM on VCE_TEMPLATE above.
 use constant KEY_TEMPLATE => 'n*';
@@ -107,6 +109,33 @@ sub UCA_Version { "9" }
 
 sub Base_Unicode_Version { $UNICODE_VERSION || 'unknown' }
 
+######
+
+use constant UNICODE_FOR_PACK => ("A" eq pack('U', 0x41));
+use constant NATIVE_FOR_PACK  => ("A" eq pack('U', ord("A")));
+
+use constant UNICODE_FOR_UNPACK => (0x41 == unpack('U', "A"));
+use constant NATIVE_FOR_UNPACK  => (ord("A") == unpack('U', "A"));
+
+sub pack_U {
+    return UNICODE_FOR_PACK
+	? pack('U*', @_)
+	: NATIVE_FOR_PACK
+	    ? pack('U*', map utf8::unicode_to_native($_), @_)
+	    : die "$PACKAGE, a Unicode code point cannot be stringified.\n";
+}
+
+sub unpack_U {
+    return UNICODE_FOR_UNPACK
+	? unpack('U*', shift)
+	: NATIVE_FOR_UNPACK
+	    ? map(utf8::native_to_unicode($_), unpack 'U*', shift)
+	    : die "$PACKAGE, a code point returned from unpack U " .
+		"cannot be converted into Unicode.\n";
+}
+
+######
+
 my (%AlternateOK);
 @AlternateOK{ qw/
     blanked  non-ignorable  shifted  shift-trimmed
@@ -119,12 +148,15 @@ our @ChangeOK = qw/
   /;
 
 our @ChangeNG = qw/
-    entry entries table combining maxlength
+    entry entries table maxlength
     ignoreChar ignoreName undefChar undefName
     versionTable alternateTable backwardsTable forwardsTable rearrangeTable
-    derivCode normCode rearrangeHash isShift L3ignorable
+    derivCode normCode rearrangeHash L3_ignorable
+    backwardsFlag
   /;
-# The hash key 'ignored' is deleted at VERSION 0.21.
+# The hash key 'ignored' is deleted at v 0.21.
+# The hash key 'isShift' is deleted at v 0.23.
+# The hash key 'combining' is deleted at v 0.24.
 
 my (%ChangeOK, %ChangeNG);
 @ChangeOK{ @ChangeOK } = ();
@@ -148,12 +180,18 @@ sub change {
     return wantarray ? %old : $self;
 }
 
+sub _checkLevel {
+    my $level = shift;
+    my $key   = shift;
+    croak sprintf "Illegal level %d (in \$self->{%s}) lower than %d.",
+	$level, $key, MinLevel if MinLevel > $level;
+    croak sprintf "Unsupported level %d (in \$self->{%s}) higher than %d ",
+	$level, $key, MaxLevel if MaxLevel < $level;
+}
+
 sub checkCollator {
     my $self = shift;
-    croak "Illegal level lower than 1 (passed $self->{level})."
-	if $self->{level} < 1;
-    croak "A level higher than 4 (passed $self->{level}) is not supported."
-	if 4 < $self->{level};
+    _checkLevel($self->{level}, "level");
 
     $self->{derivCode} =
 	$self->{UCA_Version} ==  8 ? \&_derivCE_8 :
@@ -164,13 +202,24 @@ sub checkCollator {
     croak "$PACKAGE unknown alternate tag name: $self->{alternate}"
 	unless exists $AlternateOK{ $self->{alternate} };
 
-    $self->{isShift} = $self->{alternate} eq 'shifted' ||
-		$self->{alternate} eq 'shift-trimmed';
-
-    $self->{backwards} = []
-	if ! defined $self->{backwards};
-    $self->{backwards} = [ $self->{backwards} ]
-	if ! ref $self->{backwards};
+    if (! defined $self->{backwards}) {
+	$self->{backwardsFlag} = 0;
+    }
+    elsif (! ref $self->{backwards}) {
+	_checkLevel($self->{backwards}, "backwards");
+	$self->{backwardsFlag} = 1 << $self->{backwards};
+    }
+    else {
+	my %level;
+	$self->{backwardsFlag} = 0;
+	for my $b (@{ $self->{backwards} }) {
+	    _checkLevel($b, "backwards");
+	    $level{$b} = 1;
+	}
+	for my $v (sort keys %level) {
+	    $self->{backwardsFlag} += 1 << $v;
+	}
+    }
 
     $self->{rearrange} = []
 	if ! defined $self->{rearrange};
@@ -219,7 +268,7 @@ sub new
 	$self->parseEntry($_) foreach split /\n/, $self->{entry};
     }
 
-    $self->{level} ||= 4;
+    $self->{level} ||= MaxLevel;
     $self->{UCA_Version} ||= UCA_Version();
 
     $self->{overrideHangul} = ''
@@ -301,10 +350,7 @@ sub parseEntry
     $entry = join(CODE_SEP, @uv); # in JCPS
 
     if (defined $self->{undefChar} || defined $self->{ignoreChar}) {
-	# Do not use UTF_TEMPLATE; Perl' RE is only for utf8.
-	my $ele = $IsEBCDIC
-	    ? pack('U*', map utf8::unicode_to_native($_), @uv)
-	    : pack('U*', @uv);
+	my $ele = pack_U(@uv);
 
 	# regarded as if it were not entried in the table
 	return
@@ -319,45 +365,31 @@ sub parseEntry
     $k = '[.0000.0000.0000.0000]'
 	if defined $self->{ignoreName} && $name =~ /$self->{ignoreName}/;
 
-    my $combining = TRUE; # primary = 0, secondary != 0;
-    my $level3ignore;
+    my $is_L3_ignorable;
 
     foreach my $arr ($k =~ /\[([^\[\]]+)\]/g) { # SPACEs allowed
 	my $var = $arr =~ /\*/; # exactly /^\*/ but be lenient.
 	my @wt = _getHexArray($arr);
 	push @key, pack(VCE_TEMPLATE, $var, @wt);
-	$combining = FALSE
-	    unless $wt[0] == 0 && $wt[1] != 0;
-	$level3ignore = TRUE
-	    if $wt[0] == 0 && $wt[1] == 0 && $wt[2] == 0;
+	$is_L3_ignorable = TRUE
+	    if $wt[0] + $wt[1] + $wt[2] == 0;
+	  # if $arr !~ /[1-9A-Fa-f]/; NG
+	  # Conformance Test shows L3-ignorable is completely ignorable.
     }
 
     $self->{entries}{$entry} = \@key;
 
-    $self->{combining}{$entry} = TRUE
-	if $combining;
-
-    # A Level 3 ignorable (tertiary ignorable) is regarded
-    # as a completely ignorable
-    # since this module does not support the 5th level.
-    # Variable entries (marked with *) are not included.
-
-    # The difference between [.0000.0000.0000.0000] (e.g. NULL, DELETE)
-    # and [.0000.0000.0000.Non-Zero] (e.g. HEBREW ACCENT ETNAHTA)
-    # is neglected.
-
-    # The key is a string representing a numeral code point.
-
-    $self->{L3ignorable}{$uv[0]} = TRUE
-	if @uv == 1 && $level3ignore;
+    $self->{L3_ignorable}{$uv[0]} = TRUE
+	if @uv == 1 && $is_L3_ignorable;
 
     # Contraction is to be considered in the range of this maxlength.
     $self->{maxlength}{$uv[0]} = scalar @uv
 	if @uv > 1;
 }
 
+
 ##
-## arrayref[weights] = altCE(bool variable?, list[num] weights)
+## arrayref[weights] = altCE(VCE)
 ##
 sub altCE
 {
@@ -365,26 +397,29 @@ sub altCE
     my($var, @wt) = unpack(VCE_TEMPLATE, shift);
 
     $self->{alternate} eq 'blanked' ?
-	$var ? [0,0,0,$wt[3]] : \@wt :
+	$var ? [Var1Wt, 0, 0, $wt[3]] : \@wt :
     $self->{alternate} eq 'non-ignorable' ?
 	\@wt :
     $self->{alternate} eq 'shifted' ?
-	$var ? [0,0,0,$wt[0] ]
-	     : [ @wt[0..2], $wt[0]+$wt[1]+$wt[2] ? Shift4 : 0 ] :
+	$var ? [Var1Wt, 0, 0, $wt[0] ]
+	     : [ @wt[0..2], $wt[0]+$wt[1]+$wt[2] ? Shift4Wt : 0 ] :
     $self->{alternate} eq 'shift-trimmed' ?
-	$var ? [0,0,0,$wt[0] ] : [ @wt[0..2], 0 ] :
+	$var ? [Var1Wt, 0, 0, $wt[0] ] : [ @wt[0..2], 0 ] :
         croak "$PACKAGE unknown alternate name: $self->{alternate}";
 }
 
 sub viewSortKey
 {
     my $self = shift;
-    my $ver = $self->{UCA_Version};
+    $self->visualizeSortKey($self->getSortKey(@_));
+}
 
-    my $key  = $self->getSortKey(@_);
-    my $view = join " ", map sprintf("%04X", $_), unpack(KEY_TEMPLATE, $key);
+sub visualizeSortKey
+{
+    my $self = shift;
+    my $view = join " ", map sprintf("%04X", $_), unpack(KEY_TEMPLATE, shift);
 
-    if ($ver <= 8) {
+    if ($self->{UCA_Version} <= 8) {
 	$view =~ s/ ?0000 ?/|/g;
     } else {
 	$view =~ s/\b0000\b/|/g;
@@ -407,7 +442,7 @@ sub splitCE
     my $ent  = $self->{entries};
     my $max  = $self->{maxlength};
     my $reH  = $self->{rearrangeHash};
-    my $L3i  = $self->{L3ignorable};
+    my $ign  = $self->{L3_ignorable};
     my $ver9 = $self->{UCA_Version} > 8;
 
     my ($str, @buf);
@@ -426,9 +461,7 @@ sub splitCE
     }
 
     # get array of Unicode code point of string.
-    my @src = $IsEBCDIC
-	? map(utf8::native_to_unicode($_), unpack UTF_TEMPLATE, $str)
-	: unpack(UTF_TEMPLATE, $str);
+    my @src = unpack_U($str);
 
     # rearrangement:
     # Character positions are not kept if rearranged,
@@ -445,7 +478,7 @@ sub splitCE
     if ($ver9) {
 	# To remove a character marked as a completely ignorable.
 	for (my $i = 0; $i < @src; $i++) {
-	    $src[$i] = undef if $L3i->{ $src[$i] };
+	    $src[$i] = undef if $ign->{ $src[$i] };
 	}
     }
 
@@ -483,6 +516,13 @@ sub splitCE
 	    }
 	}
 
+	if ($wLen) {
+	    for (my $p = $i + 1; $p < @src; $p++) {
+		last if defined $src[$p];
+		$i = $p;
+	    }
+	}
+
 	push @buf, $wLen ? [$ce, $i_orig, $i + 1] : $ce;
     }
     return \@buf;
@@ -525,7 +565,7 @@ sub getWt
 	    $cjk
 		? map(pack(VCE_TEMPLATE, NON_VAR, @$_), &$cjk($u))
 		: defined $cjk && $self->{UCA_Version} <= 8 && $u < 0x10000
-		    ? pack(VCE_TEMPLATE, NON_VAR, $u, Min2, Min3, $u)
+		    ? pack(VCE_TEMPLATE, NON_VAR, $u, Min2Wt, Min3Wt, $u)
 		    : $der->($u);
     }
     else {
@@ -543,21 +583,17 @@ sub getSortKey
     my $lev  = $self->{level};
     my $rCE  = $self->splitCE(shift); # get an arrayref of JCPS
     my $ver9 = $self->{UCA_Version} > 8;
-    my $sht  = $self->{isShift};
+    my $v2i  = $self->{alternate} ne 'non-ignorable';
 
     # weight arrays
     my (@buf, $last_is_variable);
 
     foreach my $wt (map $self->getWt($_), @$rCE) {
-	if ($sht && $ver9) {
-	    if ($wt->[0] == 0) {
-		if ($wt->[1] == 0 && $wt->[2] == 0) {
-		    $last_is_variable = TRUE;
-		} else {
-		    next if $last_is_variable;
-		}
+	if ($v2i && $ver9) {
+	    if ($wt->[0] == 0) { # ignorable
+		next if $last_is_variable;
 	    } else {
-		$last_is_variable = FALSE;
+		$last_is_variable = ($wt->[0] == Var1Wt);
 	    }
 	}
 	push @buf, $wt;
@@ -567,13 +603,10 @@ sub getSortKey
     my @ret = ([],[],[],[]);
     foreach my $v (0..$lev-1) {
 	foreach my $b (@buf) {
-	    push @{ $ret[$v] }, $b->[$v] if $b->[$v];
+	    push @{ $ret[$v] }, $b->[$v]
+		if 0 < $b->[$v];
 	}
     }
-    foreach (@{ $self->{backwards} }) {
-	my $v = $_ - 1;
-	@{ $ret[$v] } = reverse @{ $ret[$v] };
-    }
 
     # modification of tertiary weights
     if ($self->{upper_before_lower}) {
@@ -590,6 +623,15 @@ sub getSortKey
 	    elsif (0x0D <= $_ && $_ <= 0x0E) { $_ += 5 } # hiragana
 	}
     }
+
+    if ($self->{backwardsFlag}) {
+	for (my $v = MinLevel; $v <= MaxLevel; $v++) {
+	    if ($self->{backwardsFlag} & (1 << $v)) {
+		@{ $ret[$v-1] } = reverse @{ $ret[$v-1] };
+	    }
+	}
+    }
+
     join LEVEL_SEP, map pack(KEY_TEMPLATE, @$_), @ret;
 }
 
@@ -629,8 +671,8 @@ sub _derivCE_9 {
     my $aaaa = $base + ($u >> 15);
     my $bbbb = ($u & 0x7FFF) | 0x8000;
     return
-	pack(VCE_TEMPLATE, NON_VAR, $aaaa, Min2, Min3, $u),
-	pack(VCE_TEMPLATE, NON_VAR, $bbbb,    0,    0, $u);
+	pack(VCE_TEMPLATE, NON_VAR, $aaaa, Min2Wt, Min3Wt, $u),
+	pack(VCE_TEMPLATE, NON_VAR, $bbbb,      0,      0, $u);
 }
 
 sub _derivCE_8 {
@@ -675,20 +717,6 @@ sub _isNonCharacter {
 }
 
 
-# DEBUG for index()
-sub _viewGrapheme {
-    my $str = shift;
-    my $ret = '';
-    for my $gr (@$str) {
-	for my $wt (@$gr) {
-	    $ret .= sprintf "[%s]",
-		join '.', map sprintf("%04X", $_), @$wt;
-	}
-	$ret .= ", ";
-    }
-    return $ret;
-};
-
 ##
 ## bool _nonIgnorAtLevel(arrayref weights, int level)
 ##
@@ -697,7 +725,7 @@ sub _nonIgnorAtLevel($$)
     my $wt = shift;
     return if ! defined $wt;
     my $lv = shift;
-    return grep($wt->[$_] != 0, 0..$lv-1) ? TRUE : FALSE;
+    return grep($wt->[$_-1] != 0, MinLevel..$lv) ? TRUE : FALSE;
 }
 
 ##
@@ -744,10 +772,9 @@ sub index
        $pos   = 0 if $pos < 0;
     my $grob  = shift;
 
-    my $comb  = $self->{combining};
     my $lev   = $self->{level};
     my $ver9  = $self->{UCA_Version} > 8;
-    my $sht   = $self->{isShift};
+    my $v2i   = $self->{alternate} ne 'non-ignorable';
 
     if (! @$subCE) {
 	my $temp = $pos <= 0 ? 0 : $len <= $pos ? $len : $pos;
@@ -769,21 +796,18 @@ sub index
     for my $wt (map $self->getWt($_), @$subCE) {
 	my $to_be_pushed = _nonIgnorAtLevel($wt,$lev);
 
-	if ($sht && $ver9) {
+	if ($v2i && $ver9) {
 	    if ($wt->[0] == 0) {
-		if ($wt->[1] == 0 && $wt->[2] == 0) {
-		    $last_is_variable = TRUE;
-		} else {
-		    $to_be_pushed = FALSE if $last_is_variable;
-		}
+		$to_be_pushed = FALSE if $last_is_variable;
 	    } else {
-		$last_is_variable = FALSE;
+		$last_is_variable = ($wt->[0] == Var1Wt);
 	    }
 	}
 
-	if (@subWt && $wt->[0] == 0 && $wt->[1] != 0) {
+	if (@subWt && $wt->[0] == 0) {
 	    push @{ $subWt[-1] }, $wt if $to_be_pushed;
 	} else {
+	    $wt->[0] = 0 if $wt->[0] == Var1Wt;
 	    push @subWt, [ $wt ];
 	}
     }
@@ -796,26 +820,24 @@ sub index
     for (my $i = 0; $i <= $end; ) { # no $i++
 	my $found_base = 0;
 
+	# fetch a grapheme
 	while ($i <= $end && $found_base == 0) {
 	    for my $wt ($self->getWt($strCE->[$i][0])) {
 		my $to_be_pushed = _nonIgnorAtLevel($wt,$lev);
 
-		if ($sht && $ver9) {
+		if ($v2i && $ver9) {
 		    if ($wt->[0] == 0) {
-			if ($wt->[1] == 0 && $wt->[2] == 0) {
-			    $last_is_variable = TRUE;
-			} else {
-			    $to_be_pushed = FALSE if $last_is_variable;
-			}
+			$to_be_pushed = FALSE if $last_is_variable;
 		    } else {
-			$last_is_variable = FALSE;
+			$last_is_variable = ($wt->[0] == Var1Wt);
 		    }
 		}
 
-		if (@strWt && $wt->[0] == 0 && $wt->[1] != 0) {
+		if (@strWt && $wt->[0] == 0) {
 		    push @{ $strWt[-1] }, $wt if $to_be_pushed;
 		    $finPos[-1] = $strCE->[$i][2];
 		} elsif ($to_be_pushed) {
+		    $wt->[0] = 0 if $wt->[0] == Var1Wt;
 		    push @strWt,  [ $wt ];
 		    push @iniPos, $found_base ? NOMATCHPOS : $strCE->[$i][1];
 		    $finPos[-1] = NOMATCHPOS if $found_base;
@@ -827,6 +849,7 @@ sub index
 	    $i++;
 	}
 
+	# try to match
 	while ( @strWt > @subWt || (@strWt == @subWt && $i > $end) ) {
 	    if ($iniPos[0] != NOMATCHPOS &&
 		    $finPos[$#subWt] != NOMATCHPOS &&
@@ -863,11 +886,10 @@ sub match
 {
     my $self = shift;
     if (my($pos,$len) = $self->index($_[0], $_[1])) {
-	return wantarray
-		?  substr($_[0], $pos, $len)
-		: \substr($_[0], $pos, $len);
-	  # This is an lvalue, but subst() is more user-friendly
-	  # than ${ match() } = $replacement;
+	my $temp = substr($_[0], $pos, $len);
+	return wantarray ? $temp : \$temp;
+	# An lvalue ref \substr should be avoided,
+	# since its value is affected by modification of its referent.
     }
     else {
 	return;
@@ -1010,12 +1032,12 @@ which are marked with an ASTERISK in the table
 These names are case-insensitive.
 By default (if specification is omitted), 'shifted' is adopted.
 
-   'Blanked'        Variable elements are ignorable at levels 1 through 3;
+   'Blanked'        Variable elements are made ignorable at levels 1 through 3;
                     considered at the 4th level.
 
    'Non-ignorable'  Variable elements are not reset to ignorable.
 
-   'Shifted'        Variable elements are ignorable at levels 1 through 3
+   'Shifted'        Variable elements are made ignorable at levels 1 through 3
                     their level 4 weight is replaced by the old level 1 weight.
                     Level 4 weight for Non-Variable elements is 0xFFFF.
 
@@ -1235,9 +1257,9 @@ If the tag is made true, this is reversed.
 
 B<NOTE>: These tags simplemindedly assume
 any lowercase/uppercase or hiragana/katakana distinctions
-should occur in level 3, and their weights at level 3
-should be same as those mentioned in 7.3.1, UTS #10.
-If you define your collation elements which violates this,
+must occur in level 3, and their weights at level 3
+must be same as those mentioned in 7.3.1, UTS #10.
+If you define your collation elements which violate this requirement,
 these tags don't work validly.
 
 =back
@@ -1310,11 +1332,15 @@ and get the result of the comparison of the strings using UCA.
 
 B<DISCLAIMER:> If C<preprocess> or C<normalization> tag is true
 for C<$Collator>, calling these methods (C<index>, C<match>, C<gmatch>,
-C<subst>, C<gsubst>) are croaked,
+C<subst>, C<gsubst>) is croaked,
 as the position and the length might differ
 from those on the specified string.
 (And the C<rearrange> tag is neglected.)
 
+The C<match>, C<gmatch>, C<subst>, C<gsubst> methods work
+like C<m//>, C<m//g>, C<s///>, C<s///g>, respectively,
+but they are not aware of any pattern, but only a literal substring.
+
 =over 4
 
 =item C<$position = $Collator-E<gt>index($string, $substring[, $position])>
@@ -1391,7 +1417,7 @@ the first occurrence of the matching part is replaced by C<$replacement>
 C<$replacement> can be a C<CODEREF>,
 taking the matching part as an argument,
 and returning a string to replace the matching part
-(a bit silmilar to C<s/(..)/$coderef-E<gt>($1)/e>).
+(a bit similar to C<s/(..)/$coderef-E<gt>($1)/e>).
 
 =item C<$count = $Collator-E<gt>gsubst($string, $substring, $replacement)>
 
@@ -1402,7 +1428,7 @@ all the occurrences of the matching part is replaced by C<$replacement>
 C<$replacement> can be a C<CODEREF>,
 taking the matching part as an argument,
 and returning a string to replace the matching part
-(a bit silmilar to C<s/(..)/$coderef-E<gt>($1)/eg>).
+(a bit similar to C<s/(..)/$coderef-E<gt>($1)/eg>).
 
 e.g.
 
